// lcd.c
// LCD game framework runtime library
// Link with the user's code to use LCD library functions.

#include <stdlib.h>
#include <string.h>
#include <EzGraph.h>
#include "lcd.h"

int LCD_get_fragment_num(void); //  generated in <app>.h

static int fragment_num;
static int body_width, body_height;
static int client_left, client_top;
static int client_right, client_bottom;
static int client_width, client_height;
static LCD_fragment_t *LCD_fragments;
static LCD_unit_t *unit_queue_head = NULL;
static LCD_unit_t *unit_queue_tail = NULL;
static int unit_num = 0;

static void timerh(void);
static void keyh(int n);
//static void key_releaseh(int n);
static void check_keymap(void);
static void update_button_stats(void);
static void draw_screen(void);
static void read_xpm(void);
static void enqueue_unit(LCD_unit_t *unit);
//static LCD_unit_t *dequeue_unit(void);
static void callback_handlers_act(void);
static void callback_handlers_erase(void);
static void callback_handlers_draw(void);

typedef struct {
    LCD_button_t id;
    int key;
    int keycode;
    int x, y;
} button_info_t;

static button_info_t button_infos[] =
{
 {LCD_button_down,  EZ_Down,  EZ_Down,    72, 508},
 {LCD_button_left,  EZ_Left,  EZ_Left,    36, 472},
 {LCD_button_right, EZ_Right, EZ_Right,  108, 472},
 {LCD_button_up,    EZ_Up,    EZ_Up,      72, 436},
 {LCD_button_a,     ' ',      EZ_space, -108, 472},
 {LCD_button_b,     13,       EZ_Return, -42, 472},
};                                

typedef struct {
    int is_down;            // 現在押されていたら真
    int is_pushed;          // 押したときに一度だけ真(リピート無効)
    int is_push_checked;    // pushedフラグを立てたらupまで真
} button_stat_t;

static button_stat_t button_stats[LCD_button_end];

/*
static int keycodes[] =
{
    EZ_Down, EZ_Left, EZ_Right, EZ_Up,
    EZ_space, EZ_Return,
};
*/

static EzImage bg_img;
static int is_bg_set;
static int is_paused;

static void (*draw_score_handler)(int, int);


int LCD_init(int width, int height){
    client_width = width;
    client_height = height;
    client_left = BODY_LEFT_WIDTH;
    client_top = BODY_TOP_HEIGHT;
    client_right = client_left + client_width - 1;
    client_bottom = client_top + client_height - 1;
    body_width
        = BODY_LEFT_WIDTH + client_width + BODY_RIGHT_WIDTH;    
    body_height
        = BODY_TOP_HEIGHT + client_height + BODY_BOTTOM_HEIGHT;
    EzOpenWindow(body_width, body_height);
    
    fragment_num = LCD_get_fragment_num();
    printf("lcd.c fragment_num:%d\n", fragment_num);

    LCD_fragments = malloc(sizeof(LCD_fragment_t)*fragment_num);
    if (LCD_fragments == NULL){
        printf("malloc error!\n"); exit(1);
    }
    read_xpm();

    EzSetKeyHandler(keyh);
    //    EzSetKeyReleaseHandler(key_releaseh);
    EzSetTimerHandler(timerh, LCD_TIMER_MS);

    printf("body:%d x %d client:%d x %d\n", body_width, body_height, client_width, client_height);
    
    return 0;
}

int LCD_finalize(void){
    return 0;
}

LCD_unit_t *LCD_create_unit(int size){
    LCD_unit_t *unit;
    unit = malloc(size);
    unit->on_act = NULL;
    unit->on_erase = NULL;
    unit->on_draw = NULL;
    enqueue_unit(unit);
    return unit;
}


void LCD_show_fragment(int fragment_no){
    LCD_fragments[fragment_no].is_visible = 1;
}
void LCD_hide_fragment(int fragment_no){
    LCD_fragments[fragment_no].is_visible = 0;
}
int LCD_is_button_down(int no){
    return button_stats[no].is_down;
}
int LCD_is_button_pushed(int no){
    return button_stats[no].is_pushed;
}

void LCD_set_bg(char **xpm){
    bg_img = EzReadXPM(xpm);
    is_bg_set = 1;
}

static void timerh(void){
    check_keymap();
    callback_handlers_erase();    
    callback_handlers_act();
    callback_handlers_draw();    
    draw_screen();
    update_button_stats();
}


static void keyh(int n){
    int i;

    //    printf("key:%c(%d)\n", n, n);
    if (n == 'p'){
        if (is_paused){
            puts("PAUSE_OFF");
            EzSetTimerHandler(timerh, LCD_TIMER_MS);
            is_paused = 0;
        }
        else {
            puts("PAUSE_ON");
            EzSetTimerHandler(NULL, 0);
            is_paused = 1;
        }
    }

    for (i = 0; i < LCD_button_end; i++){
        if (n == button_infos[i].key){
            if (!button_stats[i].is_push_checked){
                button_stats[i].is_pushed = 1;
                button_stats[i].is_push_checked = 1;
            }
            printf("pushed %c(%d)\n", n, n);
            break;
        }
    }
}

static void check_keymap(void){
    int i;
    char keys[32];

    EzQueryKeymap(keys);
    for (i = 0; i < LCD_button_end; i++){
        button_stats[i].is_down
            = EzIsKeyPress(keys, button_infos[i].keycode);
        if (!button_stats[i].is_down){
            button_stats[i].is_push_checked = 0;            
        }
    }
}

static void update_button_stats(void){
    int i;    
    // is_pushedは押した直後の1フレームのみ1とするため、
    // ここで毎回0に戻す
    for (i = 0; i < LCD_button_end; i++){
        button_stats[i].is_pushed = 0;        
    }
}

static void draw_screen(void){
    int i, x, y;

    EzSetColorByRGB(120, 120, 240);
    EzFillBoxB(0, 0, body_width-1, body_height-1);
    EzSetColorByRGB(192, 192, 192);
    EzFillBoxB(BODY_BORDER_SIZE,
               BODY_BORDER_SIZE,
               body_width-1-BODY_BORDER_SIZE,
               body_height-1-BODY_BORDER_SIZE);
    //    EzSetColorByRGB(255, 255, 255);
    //    EzSetColorByRGB(180, 170, 150);
    EzSetColorByRGB(120, 120, 240);
    EzFillBoxB(client_left-SCREEN_MARGIN_SIZE-SCREEN_BORDER_SIZE,
               client_top-SCREEN_MARGIN_SIZE-SCREEN_BORDER_SIZE,
               client_right+SCREEN_MARGIN_SIZE+SCREEN_BORDER_SIZE,
               client_bottom+SCREEN_MARGIN_SIZE+SCREEN_BORDER_SIZE);
    EzSetColorByRGB(0, 0, 0);
    EzFillBoxB(client_left-SCREEN_MARGIN_SIZE,
               client_top-SCREEN_MARGIN_SIZE,
               client_right+SCREEN_MARGIN_SIZE,
               client_bottom+SCREEN_MARGIN_SIZE);
    EzSetColorByRGB(160, 156, 148);
    EzFillBoxB(client_left, client_top,
               client_right, client_bottom);


    for (i = 0; i < LCD_button_end; i++){
        x = button_infos[i].x;
        if (x < 0){
            x = body_width+x;
        }
        y = button_infos[i].y;
        if (button_stats[i].is_down){
            x+=2;
            y+=2;
            EzSetColorByRGB(120, 24, 24);
        }
        else {
            EzSetColorByRGB(192, 32, 32);
        }
        EzFillCircleB(x, y, BUTTON_R);
    }

    if (is_bg_set){
        EzPutB(client_left, client_top, bg_img);
    }
    for (i = 0; i < fragment_num; i++){
        if (LCD_fragments[i].is_visible){
            EzPutB(LCD_fragments[i].x + client_left,
                   LCD_fragments[i].y + client_top,
                   LCD_fragments[i].img);
        }
    }
    draw_score_handler(client_left, client_top);
    EzShowBuffer();
}

static void read_xpm(void){
    int i;
    
    for (i = 0; i < fragment_num; i++){
        LCD_fragments[i].img
            = EzReadXPM(LCD_fragment_infos[i].xpm);
        LCD_fragments[i].x = LCD_fragment_infos[i].x;
        LCD_fragments[i].y = LCD_fragment_infos[i].y;
        LCD_fragments[i].is_visible = 0;
    }
}

void enqueue_unit(LCD_unit_t *unit){
    if (unit_queue_tail == NULL){
        unit_queue_head = unit;
        unit_queue_tail = unit;
        unit->next = NULL;
    }
    else {
        unit_queue_tail->next = unit;
        unit->next = NULL;
        unit_queue_tail = unit;
    }
    unit_num ++;
    printf("%d Units total.\n", unit_num);
}

#ifdef RESERVE
LCD_unit_t *dequeue_unit(void){
    LCD_unit_t *unit;
    
    if (unit_queue_head == NULL){
        return NULL;
    }
    else {
        unit = unit_queue_head;
        unit_queue_head = unit->next;
    }
    unit_num --;
    return unit;
}
#endif

static void callback_handlers_act(void){
    LCD_unit_t *unit = unit_queue_head;
    while (unit != NULL){
        //        printf("calling back act of unit type %d\n", unit->type);
        if (unit->on_act != NULL){
            //            printf("called %p\n", unit->on_act);
            unit->on_act(unit);
        }
        unit = unit->next;
    }
}

static void callback_handlers_erase(void){
    LCD_unit_t *unit = unit_queue_head;
    while (unit != NULL){
        if (unit->on_erase != NULL){
            //            printf("called %p\n", unit->on_erase);
            unit->on_erase(unit);
        }
        unit = unit->next;
    }
}

static void callback_handlers_draw(void){
    LCD_unit_t *unit = unit_queue_head;
    while (unit != NULL){
        if (unit->on_draw != NULL){
            //            printf("called %p\n", unit->on_draw);
            unit->on_draw(unit);
        }
        unit = unit->next;
    }
}

void LCD_set_draw_score_handler(void(*f)(int, int)){
    draw_score_handler = f;            
}
