/*
char **xpm
プログラムにXPMファイルをincludeして得られる配列の内容を想定。
・XPMファイル中のコメント行は削除されている
・各行は文字列の中身のみを保持。
　つまり元ファイル中の行頭行末ダブルクォートおよび行末カンマは
  削除されている
  save_xpm: この形式をXPMファイルとして出力
  load_xpm: includeしていないXPMファイルを読み込んでこの形式に変換
dynamic_xpm_t
色やピクセル情報にアクセスしやすい形に変換したもの
  xpm2dynamic_xpm, dynamic_xpm2xpmで上記オンメモリxpmと
  相互変換可能


TODO
depthによって1文字か2文字で比較・コピーするのが面倒なので、
dynamic xpmのピクセル情報は1ピクセルを整数値一つで表す。
元のXPMで色に割り当てられた文字は1-2文字で、7bitコードとすると
shift+加算で合わせて14bit(16K通り)なので、雑に2文字から対応する整数値
を引くテーブルを作っても大きさはたいしたことないし、
1ピクセル4バイトでも一般的なフルカラー画像と変わらない
(実際は14ビットだから16ビット整数で足りるが、わざわざshortにする必要が
あるかどうか)

dynamic_xpm_tは現在、ピクセル情報のみ別途mallocし、
色情報は色数の最大値の数だけ決め打ちで確保している
将来的には色情報も別途mallocし、
必要な色数分だけメモリを確保する


*/



#ifndef PROCESS_XPM_H
#define PROCESS_XPM_H

// 扱える色数
#define XPM_COLOR_NUM_MAX 256
// 色名文字列の長さ上限
#define XPM_COLOR_NAME_LEN_MAX 128
// 1ピクセルに使うバイト数上限
#define DEPTH_MAX 2
// XPMファイルの1行上限
#define LINE_LEN_MAX 1024

// 1ピクセルの色を表す文字
//   - 2文字使うこともあるがnullターミネートしない
//   - 1文字の場合、chr[0]に格納しchr[1]には0を入れること
typedef struct {
    char chrs[2];
} pixel_str_t;

// 1ピクセルの色情報(pixel_str_tをエンコードした14ビットの整数値)
typedef short pixel_val_t;

// 1-2文字のピクセル情報を14ビット整数値に一対一変換
#define PixelStr2Val(chrs, depth) \
          ((depth)==1 ? (chrs)[0] \
           : (((chrs)[1] << 7)+(chrs)[0]))
#define PixelVal2Str0(val) ((val) & 0x7f)
#define PixelVal2Str1(val) (((val) >> 7) & 0x7f)

// 一色分の情報
typedef struct {
    //    pixel_str_t str; // パターン中でその色に使う文字(1,2文字)
    pixel_val_t val; // その色に使う1-2文字をエンコードした値
    char name[XPM_COLOR_NAME_LEN_MAX]; // 色名。RGB形式も含めXPMで書ける文字列
} xpm_color_t;

// XPM(char*[])を操作用に変換した形式
typedef struct {
    int width_max, height_max; // 格納可能な上限サイズ(確保領域の大きさ)
    int width, height, color_num, depth;
    //    pixel_str_t transparent_str;
    pixel_val_t transparent_val;
    xpm_color_t colors[XPM_COLOR_NUM_MAX];
    pixel_val_t *pixels;
} dynamic_xpm_t;

#define DXpmWidth(dxpm)    ((dxpm)->width)
#define DXpmHeight(dxpm)   ((dxpm)->height)
#define DXpmColorNum(dxpm) ((dxpm)->color_num)
#define DXpmDepth(dxpm)    ((dxpm)->depth)
#define DXpmTransparentVal(dxpm) ((dxpm)->transparent_val)
#define DXpmColorVal(dxpm, no) ((dxpm)->colors[no].val)
#define DXpmColorName(dxpm, no) ((dxpm)->colors[no].name)
#define DXpmPixel(dxpm, x, y) \
    ((dxpm)->pixels[(y)*((dxpm)->width)+(x)])


dynamic_xpm_t *alloc_dynamic_xpm(int width, int height);
void free_dynamic_xpm(dynamic_xpm_t *dxpmpixel);
char **alloc_xpm_buffer(int width, int height, int color_num);
void free_xpm_buffer(char **xpm);
dynamic_xpm_t * xpm2dynamic_xpm(char **xpm, dynamic_xpm_t *dxpm);
int dynamic_xpm2xpm(dynamic_xpm_t *dxpm, char **xpm);
int overlay_xpm(dynamic_xpm_t *lower_dxpm, dynamic_xpm_t *upper_dxpm, 
                int offset_x, int offset_y,
                dynamic_xpm_t *result_dxpm);
int magnify_xpm_int(int xscale, int yscale,
                    dynamic_xpm_t *src_dxpm, dynamic_xpm_t *dest_dxpm);
int magnify_xpm(double xscale, double yscale,
                dynamic_xpm_t *src_dxpm, dynamic_xpm_t *dest_dxpm);

char **load_xpm(char *path, char *name);
int save_xpm(char *path, char *name, char **xpm);

#ifdef DEBUG
void show_dynamic_xpm(dynamic_xpm_t *dxpm);
int show_xpm(char **xpm);
#endif

#endif
